from jax_fdm.goals import ScalarGoal
from jax_fdm.goals.edge import EdgeGoal


class EdgeLengthGoal(ScalarGoal, EdgeGoal):
    """
    Make an edge of a network to reach a certain length.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current edge length.
        """
        return eq_state.lengths[index]



from jax_fdm.constraints import Constraint


class NodeConstraint(Constraint):
    """
    Base class for all constraints that pertain to a node in a network.
    """
    def index_from_model(self, model):
        """
        The index of the node in a structure.
        """
        try:
            return model.structure.node_index[self.key]
        except TypeError:
            return tuple([model.structure.node_index[k] for k in self.key])



from jax.numpy import inf


class Parameter:
    """
    The base class for all parameters.

    Parameters
    ----------
    key : `int` | `Tuple[int]`
        The key in the network of the parameter.
    bound_low : `float`, optional
        The lower bound of this parameter for optimization.
        Defaults to `-inf`.
    bound_ip : `float`, optional
        The upper bound of this parameter for optimization.
        Defaults to `+inf`.
    """
    def __init__(self, key, bound_low=None, bound_up=None):
        """
        Initialize the parameter.
        """
        self.key = key
        self.attr_name = None

        self._bound_low = None
        self.bound_low = bound_low

        self._bound_up = None
        self.bound_up = bound_up

    @property
    def bound_low(self):
        """
        The lower bound of the parameter.
        """
        return self._bound_low

    @property
    def bound_up(self):
        """
        The upper bound of the parameter.
        """
        return self._bound_up

    @bound_low.setter
    def bound_low(self, value):
        if value is None:
            value = -inf
        self._bound_low = value

    @bound_up.setter
    def bound_up(self, value):
        if value is None:
            value = inf
        self._bound_up = value

    def index(self, model):
        """
        Get the index of the parameter key in the structure of a model.
        """
        raise NotImplementedError

    def value(self, model):
        """
        Get the current value of the parameter.
        """
        raise NotImplementedError


class NodeParameter(Parameter):
    """
    A node parameter.
    """
    def index(self, model):
        """
        Get the index of the key of the node parameter in the structure of a model.
        """
        return model.structure.node_index[self.key]

    def value(self, model):
        """
        Get the current value of the node parameter.
        """
        return model.structure.network.node_attribute(self.key, name=self.attr_name)


class EdgeParameter(Parameter):
    """
    An edge parameter.
    """
    def index(self, model):
        """
        Get the index of the key of the edge parameter in the structure of a model.
        """
        return model.structure.edge_index[self.key]

    def value(self, model):
        """
        Get the current value of the edge parameter.
        """
        return model.structure.network.edge_attribute(self.key, name=self.attr_name)


class EdgeForceDensityParameter(EdgeParameter):
    """
    An edge force density parameter.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "q"


class NodeAnchorParameter(NodeParameter):
    """
    A node anchor parameter.
    """
    def index(self, model):
        """
        Get the index of the key of the node parameter in the structure of a model.
        """
        return model.structure.anchor_index[self.key]


class NodeLoadParameter(NodeParameter):
    pass


class NodeAnchorXParameter(NodeAnchorParameter):
    """
    Parametrize the z coordinate of an anchor.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "x"


class NodeAnchorYParameter(NodeAnchorParameter):
    """
    Parametrize the y coordinate of an anchor.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "y"


class NodeAnchorZParameter(NodeAnchorParameter):
    """
    Parametrize the z coordinate of an anchor.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "z"


class NodeLoadXParameter(NodeLoadParameter):
    """
    Parametrize the x component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "px"


class NodeLoadYParameter(NodeLoadParameter):
    """
    Parametrize the y component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "py"


class NodeLoadZParameter(NodeLoadParameter):
    """
    Parametrize the z component of a node load.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attr_name = "pz"



"""
A gradient-based optimizer.
"""
from time import time
from itertools import groupby
from functools import partial

from jax import jit
from jax import grad

from scipy.optimize import minimize
from scipy.optimize import Bounds

from jax_fdm.parameters import ParameterManager
from jax_fdm.parameters import EdgeForceDensityParameter

from jax_fdm.optimization import Collection


# ==========================================================================
# Optimizer
# ==========================================================================

class Optimizer:
    """
    Base class for all optimizers.
    """
    def __init__(self, name, disp=True):
        self.name = name
        self.disp = disp
        self.pm = None

    def constraints(self, constraints, model, params_opt):
        """
        Returns the defined constraints in a format amenable to `scipy.minimize`.
        """
        if constraints:
            print(f"Warning! {self.name} does not support constraints. I am ignoring them.")

    def gradient(self, loss):
        """
        Compute the gradient function of a loss function.
        """
        return jit(grad(loss, argnums=0))

    def hessian(self, loss):
        """
        Compute the hessian function of a loss function.
        """
        pass

# ==========================================================================
# Loss
# ==========================================================================

    @partial(jit, static_argnums=(0, 2, 3))
    def loss(self, params_opt, loss, model):
        """
        The wrapper loss.
        """
        q, xyz_fixed, loads = self.parameters_fdm(params_opt)

        return loss(q, xyz_fixed, loads, model)

# ==========================================================================
# Goals
# ==========================================================================

    def goals(self, loss, model):
        """
        Pre-process the goals in the loss function to accelerate computations.
        """
        for term in loss.terms_error:
            goal_collections = self.collect_goals(term.goals)
            for goal_collection in goal_collections:
                goal_collection.init(model)
            term.collections = goal_collections

# ==========================================================================
# Minimization
# ==========================================================================

    def problem(self,
                model,
                loss,
                parameters=None,
                constraints=None,
                maxiter=100,
                tol=1e-6,
                callback=None):
        """
        Set up an optimization problem.
        """
        # optimization parameters
        if not parameters:
            parameters = [EdgeForceDensityParameter(edge) for edge in model.structure.edges]

        self.pm = ParameterManager(model, parameters)
        x = self.parameters_opt()

        # message
        print(f"\n***Constrained form finding***\nParameters: {len(x)} \tGoals: {loss.number_of_goals()}")

        # parameter bounds
        bounds = self.parameters_bounds()

        assert x.size == self.pm.bounds_low.size
        assert x.size == self.pm.bounds_up.size

        # build goal collections
        self.goals(loss, model)
        print(f"\tGoal collections: {loss.number_of_collections()}")

        # loss matters
        loss = partial(self.loss, loss=loss, model=model)

        print("Warming up the pressure cooker...")
        start_time = time()
        _ = loss(x)
        print(f"\tLoss warmup time: {(time() - start_time):.4} seconds")

        # gradient of the loss function
        grad_loss = self.gradient(loss)  # w.r.t. first function argument
        start_time = time()
        _ = grad_loss(x)
        print(f"\tGradient warmup time: {(time() - start_time):.4} seconds")

        # gradient of the loss function
        hessian_loss = self.hessian(loss)  # w.r.t. first function argument
        if hessian_loss:
            start_time = time()
            _ = hessian_loss(x)
            print(f"\tHessian warmup time: {(time() - start_time):.4} seconds")

        # constraints
        constraints = constraints or []
        if constraints:
            start_time = time()
            constraints = self.constraints(constraints, model, x)
            print(f"\tConstraints warmup time: {round(time() - start_time, 4)} seconds")

        opt_kwargs = {"fun": loss,
                      "jac": grad_loss,
                      "hess": hessian_loss,
                      "method": self.name,
                      "x0": x,
                      "tol": tol,
                      "bounds": bounds,
                      "constraints": constraints,
                      "callback": callback,
                      "options": {"maxiter": maxiter, "disp": self.disp}}

        return opt_kwargs

    def solve(self, opt_problem):
        """
        Solve an optimization problem by minimizing a loss function via gradient descent.
        """
        print(f"Optimization with {self.name} started...")
        start_time = time()

        # minimize
        res_q = self._minimize(opt_problem)

        print(res_q.message)
        print(f"Final loss in {res_q.nit} iterations: {res_q.fun}")
        print(f"Optimization elapsed time: {time() - start_time} seconds")

        return res_q.x

    @staticmethod
    def _minimize(opt_problem):
        """
        Scipy backend method to minimize a loss function.
        """
        return minimize(**opt_problem)

# ==========================================================================
# Parameters
# ==========================================================================

    def parameters_bounds(self):
        """
        Return a tuple of arrays with the upper and the lower bounds of optimization parameters.
        """
        return Bounds(lb=self.pm.bounds_low, ub=self.pm.bounds_up)

    def parameters_opt(self):
        """
        Return a flat array with the optimization parameters.
        """
        return self.pm.parameters_opt

    def parameters_frozen(self):
        """
        Return a flat array with the parameters that must stay constant during optimization.
        """
        return self.pm.parameters_frozen

    def parameters_fdm(self, params_opt):
        """
        Reconstruct the force density parameters from the optimization parameters.
        """
        return self.pm.parameters_fdm(params_opt)

# ==========================================================================
# Goal collections
# ==========================================================================

    @staticmethod
    def collect_goals(goals):
        """
        Convert a list of goals into a list of goal collections.
        """
        goals = sorted(goals, key=lambda g: type(g).__name__)
        groups = groupby(goals, lambda g: type(g))

        collections = []
        for _, group in groups:
            collection = Collection(list(group))
            collections.append(collection)

        return collections



import matplotlib.pyplot as plt

from jax_fdm.equilibrium import EquilibriumModel


__all__ = ["LossPlotter"]


class LossPlotter:
    """
    Plot a loss function.
    """
    def __init__(self, loss, network, *args, **kwargs):
        self.loss = loss
        self.network = network
        self.fig = plt.figure(**kwargs)

    def plot(self, xs):
        """
        Plot the loss function on an array of force densities.
        """
        print("Plotting loss function...")
        model = EquilibriumModel(self.network)

        for loss_term in [self.loss] + list(self.loss.terms):
            errors = []
            for x in xs:
                q, xyz_fixed, loads = x
                eqstate = model(q, xyz_fixed, loads)
                try:
                    error = loss_term(eqstate)
                except TypeError:
                    error = loss_term(q, xyz_fixed, loads, model)
                errors.append(error)
            plt.plot(errors, label=loss_term.name)

        plt.xlabel("Optimization iterations")
        plt.ylabel("Loss")
        plt.yscale("log")
        plt.grid()
        plt.legend()
        plt.show()

    def show(self):
        """
        Display the plot.
        """
        plt.show()



import numpy as np

from compas.data import Data

import jax.tree_util as jtu


# ==========================================================================
# Recorder
# ==========================================================================

class OptimizationRecorder(Data):
    def __init__(self, optimizer=None):
        self.optimizer = optimizer
        self.history = []

    def record(self, value):
        self.history.append(value)

    def __call__(self, xk, *args, **kwargs):
        if self.optimizer:
            xk = self.optimizer.parameters_fdm(xk)
        self.record(xk)

    @property
    def data(self):
        data = {}
        data["history"] = jtu.tree_map(lambda leaf: np.asarray(leaf, dtype=np.float64).tolist(), self.history)
        return data

    @data.setter
    def data(self, data):
        self.history = data["history"]
